import traceback

from django.db.backends.utils import logger

from cavalry.locals import get_storage
from cavalry.stack import Stack


def record(*, db, duration, params, sql, times=None, error=None):
    if get_storage().get('db_record_stacks'):
        # The two last frames are in cavalry, so slice them off
        stack = Stack(traceback.extract_stack()[:-2])
    else:
        stack = []
    db.queries_log.append(
        {
            'sql': (f'{times} times: {sql}' if times else sql),
            'time': f"{duration:.3f}",
            'hrtime': duration,
            'stack': stack,
            'error': error,
        }
    )
    logger.debug(
        '(%.3f) %s; args=%s',
        duration,
        sql,
        params,
        extra={'duration': duration, 'sql': sql, 'params': params},
    )



from contextlib import contextmanager

import django
from django.db import connections
from django.db.backends import utils as db_backend_utils
from django.db.backends.utils import CursorWrapper

from cavalry.db_common import record
from cavalry.timing import get_time

assert django.VERSION[0] == 2, "This module won't work with Django 3"

DefaultCursorDebugWrapper = db_backend_utils.CursorDebugWrapper


class CavalryCursorDebugWrapper(CursorWrapper):
    # This is an enhanced version of `django.db.backends.utils.CursorDebugWrapper`.

    def execute(self, sql, params=None):
        start = get_time()
        try:
            return super().execute(sql, params)
        finally:
            stop = get_time()
            duration = stop - start
            sql = self.db.ops.last_executed_query(self.cursor, sql, params)
            self._record(sql, params, duration)

    def executemany(self, sql, param_list):
        start = get_time()
        try:
            return super().executemany(sql, param_list)
        finally:
            stop = get_time()
            duration = stop - start
            try:
                times = len(param_list)
            except TypeError:  # param_list could be an iterator
                times = '?'
            self._record(sql, param_list, duration, times)

    def _record(self, sql: str, params, duration: float, times=None):
        record(db=self.db, sql=sql, params=params, duration=duration, times=times)


@contextmanager
def enable_db_tracing():
    for conn in connections.all():
        db_backend_utils.CursorDebugWrapper = CavalryCursorDebugWrapper
        conn._cavalry_old_force_debug_cursor = conn.force_debug_cursor
        conn.force_debug_cursor = True
    yield
    for conn in connections.all():
        conn.force_debug_cursor = conn._cavalry_old_force_debug_cursor
        db_backend_utils.CursorDebugWrapper = DefaultCursorDebugWrapper



import json

import pytest
from django.test import Client
import requests_mock


@pytest.mark.django_db
@pytest.mark.parametrize('enable', (False, True))
@pytest.mark.parametrize('as_admin', (False, True))
@pytest.mark.parametrize('posting', (False, True))
def test_cavalry(settings, as_admin, enable, posting, admin_user):
    settings.CAVALRY_ENABLED = enable
    settings.CAVALRY_ELASTICSEARCH_URL_TEMPLATE = (
        'http://localhost:59595/asdf/foo' if posting else None
    )
    client = Client()
    if as_admin:
        user = admin_user
        client.force_login(user)
    else:
        user = None
    with requests_mock.mock() as m:
        m.post(
            settings.CAVALRY_ELASTICSEARCH_URL_TEMPLATE,
            status_code=201,
            json={'ok': True},
        )
        content = client.get('/').content

    # Check precondition: the user seemed logged in
    assert (admin_user.username.encode() in content) == as_admin

    # Check that the injection div only appears for admins
    assert (b'<div' in content) == (enable and as_admin)

    # Check that stats are posted only when posting is enabled
    assert bool(m.called) == (enable and posting)
    if enable and posting:
        payload = json.loads(m.request_history[-1].body.decode('utf-8'))
        # Check that user IDs are sent
        assert payload.get('user_id') == (user.id if user else None)
        database_info = payload['databases']['default']
        if database_info['n_queries']:
            assert database_info['time'] > 0



import json
from typing import List, Optional

from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.utils.crypto import get_random_string

from cavalry.policy import can_report_stacks
from cavalry.stack import Stack

STYLE = '''
#cv {
position:fixed;
z-index:9999;
bottom:0;
left:50%;
transform: translateX(-50%);
text-align:center;
font:10px/1.1 menlo,consolas,monospace;
background:rgba(0,0,0,0.8);
color:#fff;
padding:1px;
}
'''.replace(
    '\n', ''
).strip()


def inject_html(
    request: WSGIRequest, response: HttpResponse, data: dict, summary_data: dict
) -> None:
    try:
        body_closing_index = response.content.rindex(b'</body>')
    except ValueError:
        return
    content = ' | '.join(
        f'{key}={round(value, 3) if isinstance(value, float) else value}'
        for (key, value) in sorted(summary_data.items())
    )
    ns = f'cv_{get_random_string()}'
    html = f"<style>{STYLE.replace('#cv', '#' + ns)}</style><div id=\"{ns}\">{content}</div>"
    script = generate_console_script(data, with_stacks=can_report_stacks(request))
    script = '\n'.join(script)
    html += f"<script>{script}</script>"

    response.content = (
        response.content[:body_closing_index]
        + html.encode('utf-8')
        + response.content[body_closing_index:]
    )

    if 'content-length' in response:
        response['content-length'] = len(response.content)


def generate_console_script(data: dict, with_stacks: bool = False) -> List[str]:
    script = []
    for db, data in data.get('databases', {}).items():
        queries = data['queries']
        header = f'{db}: {data["n_queries"]} queries ({data["time"]} msec)'
        script.append(f'console.group({json.dumps(header)});')
        for query in queries:
            stack: Optional[Stack] = query.get('stack', None) if with_stacks else None
            query_time = query.get('hrtime', 0) * 1000
            script.append(f'console.log({query_time:.3f}, {json.dumps(query["sql"])});')
            if stack:
                script.append('console.groupCollapsed("Stack");')
                stack_lines = stack.as_lines()
                stack_json = json.dumps('\n'.join(stack_lines))
                script.append(f'console.log({stack_json});')
                script.append('console.groupEnd();')

        script.append('console.groupEnd();')
    return script


def inject_stats(request: WSGIRequest, response: HttpResponse, data: dict) -> None:
    summary_data = summarize_data(data)
    content_type = response.get('content-type', '').lower()
    if content_type.startswith('text/html') and (
        'charset' not in content_type or 'utf-8' in content_type
    ):
        inject_html(request, response, data, summary_data)
    response['x-cavalry-data'] = json.dumps(summary_data)


def summarize_data(data: dict) -> dict:
    summary_data = {
        'request time': data['duration'] * 1000,
    }
    for db, db_data in data.get('databases', {}).items():
        summary_data[f'queries [{db}]'] = db_data['n_queries']
        summary_data[f'time [{db}]'] = db_data['time']
    return summary_data



from contextlib import contextmanager
from threading import local

_storage = local()


@contextmanager
def managed(**kwargs):
    _storage.data = dict(**kwargs)
    yield _storage.data
    del _storage.data


def get_storage() -> dict:
    return getattr(_storage, 'data', {})



from random import random
from typing import Optional

from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest


def get_request_es_url_template(request: WSGIRequest) -> Optional[str]:
    return getattr(settings, 'CAVALRY_ELASTICSEARCH_URL_TEMPLATE', None)


def can_cavalrize(request: WSGIRequest) -> bool:
    if not getattr(settings, 'CAVALRY_ENABLED', False):
        return False
    probability = getattr(settings, 'CAVALRY_PROBABILITY', 1.0)
    return probability >= 1 or (random() < probability)


def can_post_stats(request: WSGIRequest) -> bool:
    return bool(get_request_es_url_template(request))


def can_record_stacks(request: WSGIRequest) -> bool:
    # By default, only record stacks if we're
    # allowed to and going to use them for posting or reporting.
    if getattr(settings, 'CAVALRY_DB_RECORD_STACKS', True):
        return can_post_stats(request) or can_report_stacks(request)
    return False


def can_serialize_stacks(request: None) -> bool:
    # NB: request will always be None here
    return getattr(settings, 'CAVALRY_POST_STACKS', True)


def can_report_stacks(request: WSGIRequest) -> bool:
    return bool(request.GET.get('_cavalry_stacks'))


def can_inject_stats(request: WSGIRequest) -> bool:
    # When debugging, stats can always be shown
    if settings.DEBUG:
        return True

    # When the user (if there is one) is a superuser, stats can be shown
    user = getattr(request, 'user', None)
    if user and getattr(user, 'is_superuser', False):
        return True

    return False


def can_post_threaded(request: WSGIRequest) -> bool:
    return getattr(settings, 'CAVALRY_THREADED_POST', False)



import re
from traceback import FrameSummary
from typing import List

ELIDED_FILENAMES = [
    'socketserver.py',
    'threading.py',
    # Middleware-related stack frames
    'django/utils/deprecation.py',
    'django/core/handlers/exception.py',
]


class Stack:  # Wrapped as an object for easier serialization
    def __init__(self, frames: List[FrameSummary]) -> None:
        self.frames = frames

    def as_lines(self) -> List[str]:
        stack_lines = []
        for filename, lineno, fn, text in self.frames:
            if any(filename.endswith(elided_end) for elided_end in ELIDED_FILENAMES):
                continue
            filename = re.sub(r'^.+site-packages/', '~/', filename)
            filename = re.sub(r'^.+lib/python\d\.\d/', '!/', filename)
            stack_lines.append(f'{filename}:{lineno} ({fn}) {text}')
        return stack_lines



import json
import platform
from datetime import datetime
from logging import getLogger

import requests
from django.core.handlers.wsgi import WSGIRequest
from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpResponse
from django.utils.encoding import force_bytes
from requests import Response

from cavalry.policy import can_serialize_stacks, get_request_es_url_template
from cavalry.stack import Stack

try:
    from ipware import get_ip
except ImportError:

    def get_ip(request: WSGIRequest) -> str:
        return request.META.get('REMOTE_ADDR')


log = getLogger(__name__)

sess = requests.Session()  # Persistent requests session, yes


class PayloadJSONEncoder(DjangoJSONEncoder):
    def default(self, obj):
        if isinstance(obj, Stack):
            if not can_serialize_stacks(request=None):
                return ''
            return '\n'.join(obj.as_lines())
        return super().default(obj)


def post_stats(request: WSGIRequest, response: HttpResponse, data: dict) -> Response:
    es_url_template = get_request_es_url_template(request)
    if not es_url_template:
        return None
    payload = build_payload(data, request, response)
    es_url = es_url_template.format_map(
        dict(payload, ymd=datetime.utcnow().strftime('%Y-%m-%d')),
    )
    body = force_bytes(json.dumps(payload, cls=PayloadJSONEncoder))
    try:
        resp = sess.post(
            es_url, data=body, headers={'Content-Type': 'application/json'}, timeout=0.5
        )
        if resp.status_code != 201:
            log.warning(
                'Unable to post data to %s (error %s): %s',
                es_url,
                resp.status_code,
                resp.text,
            )
        return resp
    except Exception as e:
        log.warning('Unable to post data to %s: %s', es_url, e)


def build_payload(data: dict, request: WSGIRequest, response: HttpResponse) -> dict:
    payload = {
        'node': platform.node(),
        'content-type': response.get('content-type'),
        'ip': get_ip(request),
        'method': request.method,
        'params': dict(request.GET),
        'path': request.path,
        'status': response.status_code,
        'time': datetime.utcnow(),
    }
    resolver_match = getattr(request, 'resolver_match', None)
    if resolver_match:
        payload.update(
            {'view': resolver_match.view_name,}
        )
    user = getattr(request, 'user', None)
    if user:
        payload.update(
            {'user_id': user.id,}
        )
    payload.update(data)
    payload.pop('start_time', None)
    payload.pop('end_time', None)
    payload.pop('db_record_stacks', None)
    return payload



from threading import Thread

import django
from django.db import connections
from django.template.response import SimpleTemplateResponse

from cavalry.locals import managed
from cavalry.policy import (
    can_cavalrize,
    can_inject_stats,
    can_post_stats,
    can_post_threaded,
    can_record_stacks,
)
from cavalry.poster import post_stats
from cavalry.reporter import inject_stats
from cavalry.timing import get_time

if django.VERSION[0] == 2:
    from cavalry import db_django2 as db_module
else:
    from cavalry import db_django3 as db_module


def _process(request, get_response):
    with db_module.enable_db_tracing(), managed(
        db_record_stacks=can_record_stacks(request),
    ) as data:
        data['start_time'] = get_time()
        response = get_response(request)
        if isinstance(response, SimpleTemplateResponse):
            response.render()
        data['end_time'] = get_time()
        data['duration'] = data['end_time'] - data['start_time']
        data['databases'] = {}
        for conn in connections.all():
            queries = [q for q in conn.queries if "hrtime" in q]
            if queries:
                total_time = sum(q.get('hrtime', 0) * 1000 for q in queries)
            else:
                total_time = 0
            data['databases'][conn.alias] = {
                'queries': queries,
                'n_queries': len(queries),
                'time': total_time,
            }

    if can_inject_stats(request):
        inject_stats(request, response, data)

    if can_post_stats(request):
        post_stats_kwargs = {'request': request, 'response': response, 'data': data}
        if can_post_threaded(request):
            Thread(
                name='cavalry poster',
                target=post_stats,
                kwargs=post_stats_kwargs,
                daemon=False,
            ).start()
        else:
            post_stats(**post_stats_kwargs)

    return response


def cavalry(get_response):
    def middleware(request):
        if not can_cavalrize(request):
            return get_response(request)
        return _process(request, get_response)

    return middleware



from contextlib import contextmanager

import django
from django.db import connection

from cavalry.db_common import record
from cavalry.timing import get_time

assert django.VERSION[0] != 2, "This module won't work with Django 2"


def log_query(execute, sql, params, many, context):
    db = context['connection']
    start = get_time()
    error = None
    try:
        result = execute(sql, params, many, context)
    except Exception as e:
        error = str(e)
        raise
    else:
        return result
    finally:
        duration = get_time() - start
        record(db=db, sql=sql, params=params, duration=duration, error=error)


@contextmanager
def enable_db_tracing():
    with connection.execute_wrapper(log_query):
        yield
